/*
 *		b t r e e c o m _ c
 *
 *	This contains internal support routines that are
 *	common to the implementation of both standard binary 
 *	and balanced binary trees
 *
 *
 * (c) Copyright 1996-2000 David J. Walker
 *	   Email:  dave.walker@icl.com
 *
 *	   Permission to copy and/or distribute granted under the
 *	   following conditions:
 *
 *	   1). This notice must remain intact.
 *	   2). The author is not responsible for the consequences of use
 *		   this software, no matter how awful, even if they
 *		   arise from defects in it.
 *	   3). Altered version must not be represented as being the
 *		   original software.
 *
 *
 *	AMENDMENT HISTORY
 *	~~~~~~~~~~~~~~~~~
 *	12 May	1996  DJW - First version
 *
 *	07 July 2000  DJW - Fixed problem in removing a node pointed out by T.J.Brown
 *						The left branch was not being correctly re-attached.
 */

#include <liblist.h>
#include "liblistx.h"


/*========================================================== BTREE_LEFT */
static
_btree_t	BTREE_Left	(_btree_t node)
/*			~~~~~~~~~~
 *	This function will loop down left nodes until 
 *	we find a node that is a leaf.
 *----------------------------------------------------------------------*/
{
	while (node->left != NULL)
	{
		node = node->left;
	}
	return (node);
}

/*========================================================= BTREE_RIGHT */
static
_btree_t	BTREE_Right  (_btree_t node)
/*			~~~~~~~~~~~
 *	This function will loop down right nodes until 
 *	we find a node that is a leaf.
 *----------------------------------------------------------------------*/
{
	while (node->right != NULL)
	{
		node = node->right;
	}
	return (node);
}


/*==================================================== _LIST_BTREEFIRST */
extern
node_t	  _LIST_BtreeFirst	(list_t list)
/*		  ~~~~~~~~~~~~~~~~
 *	This function will find the first node by
 *	starting at the base and follwing left
 *	branches until a leaf is reached.
 *----------------------------------------------------------------------*/
{
	_btree_t	btreenode = (_btree_t)list->nodes;

	if (btreenode != NULL)
	{
		btreenode = BTREE_Left(btreenode);
		LIBASSERT (btreenode != NULL);
		btreenode = _LIST_USERNODE(list,btreenode);
	}
	return ((node_t)btreenode);
}


/*==================================================== _LIST_BTREELAST */
extern
node_t	  _LIST_BtreeLast  (list_t list)
/*		  ~~~~~~~~~~~~~~~
 *	This function will find the last node by
 *	starting at the base and follwing right
 *	branches until a leaf is reached.
 *----------------------------------------------------------------------*/
{
#ifdef LIBTEST
	register _btreeh_t	btreelist = (_btreeh_t)list;
#else
#define btreelist ((_btreeh_t)list)
#endif
	_btree_t	btreenode;
	
	btreenode = btreelist->nodes;

	if (btreenode != NULL)
	{
		btreenode = BTREE_Right(btreenode);
		LIBASSERT (btreenode != NULL);
		btreenode = _LIST_USERNODE(btreelist,btreenode);
	}
	return ((node_t)btreenode);
}
#undef btreelist


/*================================================== _LIST_BTREEPREVIOUS */
extern
node_t	_LIST_BtreePrevious 	 (list_t list, node_t node)
/*		~~~~~~~~~~~~~~~~~~~
 *-----------------------------------------------------------------------*/
{
	register _btree_t  leaf = (_btree_t)node;
	/*
	 *	If there is a left strand to
	 *	the current leaf, it will contain
	 *	the previous node.
	 */
	if (leaf->left)
	{
		leaf = BTREE_Right(leaf->left);
	}
	else
	{
		/*
		 *	If there is no left strand then go
		 *	upwards until we find one of which
		 *	we are not already the left branch.
		 */
		for ( ; ; )
		{
			_btree_t   oldleaf;

			oldleaf = leaf;
			leaf = oldleaf->up;
			/*
			 *	If we are already at the top level,
			 *	then there cannot be a lesser node.
			 */
			if (leaf == NULL)
			{
				return NULL;
			}
			/*
			 *	If we came from the right branch,
			 *	then this is the node that we want.
			 */
			if (leaf->right == oldleaf)
			{
				break;
			}
		}
	}
	leaf = _LIST_USERNODE(list,leaf);
	return ((node_t)leaf);
}


/*====================================================== _LIST_BTREENEXT */
extern
node_t	_LIST_BtreeNext 	 (list_t list, node_t node)
/*		~~~~~~~~~~~~~~~
 *	To find the next node, we need to either
 *	descend down the right hand path until we get to
 *	a new greatest node, or if there is no right
 *	node we must ascend through the parent links
 *	until we find a node that has a right link
 *	or we run out of nodes.
 *-----------------------------------------------------------------------*/
{
	register _btree_t leaf = (_btree_t)node;

	/*
	 *	If there is a right strand to
	 *	the current leaf, it will contain
	 *	the next node.
	 */
	if (leaf->right != NULL)
	{
	   leaf = BTREE_Left (leaf->right);
	}
	else
	{
		/*
		 *	If there is no right strand then
		 *	Go upwards until we find one of
		 *	which we are not already the 
		 *	right branch.
		 */
		for ( ; ; )
		{
			_btree_t   oldleaf;

			oldleaf = leaf;
			leaf = oldleaf->up;
			/*
			 *	If we have run out of levels,
			 *	then we are at the end of the list.
			 */
			if (leaf == NULL)
			{
				return NULL;
			}
			/*
			 *	If we came from the left branch,
			 *	then this is the node that we want.
			 */
			if (leaf->left == oldleaf)
			{
				break;
			}
		}
	}
	leaf = _LIST_USERNODE(list,leaf);
	return ((node_t)leaf);
}


/*===================================================== _LIST_BTREEREMOVE */
extern
int 	_LIST_BtreeRemove	 (list_t list, node_t node)
/*		~~~~~~~~~~~~~~~~~
 *	Remove a node from the list.
 *
 *	This is easy if it is a leaf node, 
 *	but rather more difficult if it is not.
 *-----------------------------------------------------------------------*/
{   
#ifdef LIBTEST
	register _btreeh_t btreelist = (_btreeh_t)list;
	register _btree_t  btreenode = (_btree_t)node;
#else
#define btreelist ((_btreeh_t)list)
#define btreenode ((_btree_t)node)
#endif
	register _btree_t  parent;		 /* parent node */
	register _btree_t  child;		 /* Child node to be promoted a level */

	LIBASSERT (btreelist != NULL);
	LIBASSERT (btreenode != NULL);

#ifdef LIBTEST
	if (btreenode->up == NULL)
	{
		LIBASSERT (btreelist->nodes == btreenode);
	}
#endif
	/*
	 *	We promote the right leg in
	 *	preference to the left leg.
	 */
	if (btreenode->right != NULL)
	{
		_btree_t	leftbranch;

		child = btreenode->right;
		leftbranch = btreenode->left;
		if (NULL != leftbranch)
		{
			/*
			 *	Attache nodes old left branch at correct point
			 */
			_btree_t   leftparent;

			leftparent = BTREE_Left (child);
			LIBASSERT (leftparent->left == NULL);
			leftparent->left = leftbranch;
			leftbranch->up = leftparent;
		}
	}
	else 
	{
		child = btreenode->left;
	}
	
	parent = btreenode->up;
	if (parent == NULL)
	{
		/*
		 *	The old node was the base of the
		 *	tree, so we need to set up a new
		 *	base to the tree.
		 */
		btreelist->nodes = child;
	}
	else
	{
		/*
		 *	If the old node was not the base,
		 *	then we must adjust the correct
		 *	leg of the parent to point to
		 *	its new child node.
		 */
		if (parent->right == btreenode)
		{
			parent->right = child;
		}
		else
		{
			LIBASSERT (parent->left == btreenode);
			parent->left = child;
		}
	}
	/*
	 *	Now connect the child to its new parent
	 */
	if (child != NULL)
	{
		child->up = parent;
	}

	/*
	 *	Clear any pointers from this node
	 *	so we can tell later it is not
	 *	attached to any list.
	 */
	btreenode->up = NULL;
	btreenode->right = NULL;
	btreenode->left = NULL;
	/*
	 *	Finally correct count
	 *	and do a good exit!
	 */
	return (LIST_ERROR_NONE);
}
#undef btreelist
#undef btreenode


/*====================================================== _LIST_BTREEFIND */
extern
node_t	_LIST_BtreeFind (list_t list, int (*findfunc)(node_t, va_list), va_list ap)
/*		~~~~~~~~~~~~~~~
 *	This routine walks down a binary tree to
 *	find the node that matches the desired
 *	value, or until it cannot be found.
 *-----------------------------------------------------------------------*/
{
#ifdef LIBTEST
	register _btreeh_t btreelist = (_btreeh_t)list;
#else
#define btreelist ((_btreeh_t)list)
#endif

	register _btree_t listnode;
	register node_t usernode;
	register int flag;
	register int reply;

	/*
	 *	If the list is unordered, then
	 *	search from the start.	If it is
	 *	ordered then we try and start
	 *	search (not the best way!).
	 */
	for (flag = 0, listnode = btreelist->nodes ; listnode != NULL ; )
	{
		usernode = _LIST_USERNODE(btreelist,listnode);
		reply = (*findfunc)(usernode,ap);

		if (reply == 0) 
		{
			return (usernode);
		}
		switch (flag)
		{
		case 0: 	/* Search type unknown */
				if (btreelist->listclass->compare == NULL || reply == LIST_ORDER_NA)
				{
					listnode = _LIST_REALNODE(btreelist,LIST_First((list_t)btreelist));
					flag = 2;
					break;
				}
				flag = 1;
				/* FALLTHRU */
		case 1: 	/* Search type ordered */
				listnode = (reply < 0) ? listnode->right : listnode->left;
				break;

		case 2: 	/* Search type unordered */
				listnode = _LIST_REALNODE(btreelist, _LIST_BtreeNext((list_t)btreelist,(node_t)listnode));
				break;
		}
	}

	return NULL;
}
#undef btreelist
