/*
 *	   LIST_N a m e C o m _ c
 *
 *	This file is part of the LIBLIST library.
 *	It contains the Name Space handling routines
 *	(the shareable components).
 *
 *
 * (c) Copyright 1997-2000 David J. Walker
 *	   Email:  dave.walker@icl.com
 *
 *	   Permission to copy and/or distribute is granted under the
 *	   following conditions:
 *
 *	   1). This notice must remain intact.
 *	   2). The author is not responsible for the consequences of use
 *		   this software, no matter how awful, even if they arise
 *		   from defects in it.
 *	   3). Altered versions must not be represented as being the
 *		   original software.
 *
 *	AMENDMENT HISTORY
 *	~~~~~~~~~~~~~~~~~
 *	10 May 99	DJW   - Fixed problem where name was not being set up
 *						correctly by the InitNode routine.
 *
 *	07 Jul 00	DJW   - Added code to use hashvalues.  Should speed up
 *						searching for matches slightly, particularly as
 *						the list gets larger.
 *					  - Changed list type to BTREE (was BALTREE) as the
 *						hashvalue checks should gain more than the cost
 *						of searching a slightly deeper tree, and the
 *						overheads of balancing the tree are removed.
 */

#include "liblistx.h"
#include <stdlib.h>
#include <string.h>


/***********************************************************************/
/*************************** PRIVATE FUNCTIONS *************************/
/***********************************************************************/

/*================================================== NAMESPACE_HASHVALUE
 *
 * @func	Calculate hash value for a name
 *
 * @rdesc	Result of hash calculation
 *---------------------------------------------------------------------*/
static
unsigned	NameSpace_HashValue  (
/*			~~~~~~~~~~~~~~~~~~~*/
						register char * name)	/* @param  Name to be hashed */
{
	register unsigned hashvalue = 0;

	while (*name != '\0')
	{
		hashvalue += (unsigned)*name++;
	}
	return hashvalue;
}


/*==================================================== NAMESPACE_INITNODE */
static
int 	NameSpace_InitNode	(node_t node, va_list ap)
/*		~~~~~~~~~~~~~~~~~
 *	Set up a node in the name space
 *-----------------------------------------------------------------------*/
{
#ifdef LIBTEST
	register _namespace_t pNameSpace = node;
#else
#define pNameSpace ((_namespace_t)node)
#endif /* LIBTEST */
	register char * name;

	name = va_arg(ap,char *);
	if (NULL == (pNameSpace->name = malloc (strlen(name)+1)))
	{
		return (LIST_ERROR_NOMEMORY);
	}
	(void)strcpy(pNameSpace->name, name);
	pNameSpace->count = 0;
	pNameSpace->hashvalue = (unsigned short)NameSpace_HashValue (pNameSpace->name);
	return LIST_ERROR_NONE;
}
#undef pNameSpace

/*==================================================== NAMESPACE_KILLNODE */
static
int 	NameSpace_KillNode	(node_t node)
/*		~~~~~~~~~~~~~~~~~
 *	Kill a node in the name space.
 *-----------------------------------------------------------------------*/
{
	free (((_namespace_t)node)->name);
	return LIST_ERROR_NONE;
}


/*==================================================== NAMESPACE_COMPNODE */
static
int 	NameSpace_CompNode	(node_t node1, node_t node2)
/*		~~~~~~~~~~~~~~~~~~
 *	Symbol comparison routine.
 *	
 *	We do the first comparison on the has value as it fastest.
 *	Only in the (unlikely) event that this fails do we do the string comparison).
 *-----------------------------------------------------------------------*/
{
	int 	reply;
	reply = ((_namespace_t)node1)->hashvalue - ((_namespace_t)node2)->hashvalue;
	if (reply == 0)
	{
		reply = strcmp (((_namespace_t)node1)->name,((_namespace_t)node2)->name);
	}
	return (reply);
}


/*=================================================== NAMESPACE_FINDNODE */
static
int 	NameSpace_FindNode	(node_t node, va_list ap)
/*		~~~~~~~~~~~~~~~~~~
 *	Called to try and find whether the given node is the
 *	one that is being looked for.  Called via LIST_Find().
 *
 *	We always treat NameSpace as case significant.
 *---------------------------------------------------------------*/
#define pNameSpace ((_namespace_t)node)
{
	int 	reply;
	char *	name = va_arg(ap, char *);
	unsigned short hashvalue = (unsigned short)va_arg (ap, unsigned int);

	reply = pNameSpace->hashvalue - hashvalue;
	if (reply == 0)
	{
		reply = strcmp (pNameSpace->name, name);
	}
	return (reply);
}
#undef pNameSpace


/***********************************************************************/
/**************************** PUBLIC FUNCTIONS *************************/
/***********************************************************************/


/*==================================================== NAMESPACE_USE */
extern
char *	LIST_NameSpace_Use	(char * name)
/*		~~~~~~~~~~~~~~~~~~
 *	Called to get a namespace entry.
 *	If necessary, the list and/or a
 *	new node is created.
 *-------------------------------------------------------------------*/
{
	register _namespace_t pNameSpace;

	if (_LIST_NameSpace == NULL)
	{
		_LIST_NameSpace = LIST_Create (LIST_CLASS_BTREE,
										sizeof (struct _namespace),
										NameSpace_InitNode,
										NameSpace_KillNode,
										NameSpace_CompNode);
		if (_LIST_NameSpace == NULL)
		{
			return NULL;
		}
	}
	pNameSpace = LIST_Find (_LIST_NameSpace,
							NameSpace_FindNode,
							name,
							NameSpace_HashValue(name));
	if (NULL == pNameSpace)
	{
		pNameSpace = LIST_NewAdd (_LIST_NameSpace, name);
		if (pNameSpace == NULL)
		{
			return NULL;
		}
	}
	LIBASSERT (0 == strcmp(name, pNameSpace->name));
	pNameSpace->count += 1;
	return pNameSpace->name;
}


/*===================================================== NAMESPACE_FREE */
extern
void	LIST_NameSpace_Free (char ** pname)
/*		~~~~~~~~~~~~~~~~~~~~
 *	Called when a namespace entry is freed.
 *
 *	If the count reaches zero, then we free the node
 *---------------------------------------------------------------------*/
{
	if ((pname != NULL)
	&&	(*pname != NULL))
	{
		_namespace_t pNameSpace;

		pNameSpace = LIST_Find (_LIST_NameSpace,
								NameSpace_FindNode,
								*pname,
								NameSpace_HashValue(*pname));
		if (NULL != pNameSpace)
		{
			LIBDEBUG ("LIST_NameSpace_Free", \
					"'count' parameter is not greater than 0", \
					 pNameSpace->count > 0);

			pNameSpace->count--;
			if (0 == pNameSpace->count)
			{
				(void)LIST_Delete (_LIST_NameSpace, &pNameSpace);
			}
		}
		*pname = NULL;
	}
	return ;
}
